/* rsa-int.c - Provide access to cryptographic routines through standard interface */

#include <stdio.h>
#include "rsa.h"
#include "numalgs.h"

int 
rsa_encrypt(in, out, key, parm)
	unsigned       *in;
	unsigned      **out;
	struct RSAPublicKey *key;
	struct RSAParameters *parm;
{
	int             shift;
	unsigned       *m;
	unsigned       *r;
	unsigned       *b;
	unsigned       *u;
	unsigned       *v;
#ifndef _HAVE_NO_CFREE_
	void		cfree();
#endif

	m = (unsigned *) calloc((unsigned)parm->len, sizeof(unsigned));
	r = (unsigned *) calloc((unsigned)parm->len, sizeof(unsigned));
	b = (unsigned *) calloc((unsigned)parm->len, sizeof(unsigned));
	u = (unsigned *) calloc((unsigned)parm->len, sizeof(unsigned));
	v = (unsigned *) calloc((unsigned)parm->len, sizeof(unsigned));

	shift = bitsfree(key->n, parm->len - 1);
	bsl(m, key->n, parm->len - 1, shift);
	bsl(b, in, parm->len - 1, shift);

	if (key->e == 65537)
		fermatexp(r, b, 15, m, parm->len - 1, shift, u, v);
	else
		smallexp(r, b, key->e, m, parm->len - 1, shift, u, v);

	bsr(r, r, parm->len - 1, shift);

	cfree((char*)m);
	cfree((char*)b);
	cfree((char*)u);
	cfree((char*)v);

	(*out) = r;

	return (OK);
}

int 
rsa_decrypt(in, out, key, parm)
	unsigned       *in;
	unsigned      **out;
	struct RSASecretKey *key;
	struct RSAParameters *parm;
{
	int             i;
	int             pshift;
	int             qshift;
	unsigned       *r;
	unsigned       *p1;
	unsigned       *q1;
	unsigned       *b1;
	unsigned       *b2;
	unsigned       *u;
	unsigned       *v;
	unsigned       *t1;
	unsigned      **table;
#ifndef _HAVE_NO_CFREE_
	void		cfree();
#endif

	r = (unsigned *) calloc((unsigned) parm->len, sizeof(unsigned));
	p1 = (unsigned *) calloc((unsigned) parm->len / 2, sizeof(unsigned));
	q1 = (unsigned *) calloc((unsigned) parm->len / 2, sizeof(unsigned));
	b1 = (unsigned *) calloc((unsigned) parm->len, sizeof(unsigned));
	b2 = (unsigned *) calloc((unsigned) parm->len, sizeof(unsigned));
	u = (unsigned *) calloc((unsigned) parm->len, sizeof(unsigned));
	v = (unsigned *) calloc((unsigned) parm->len, sizeof(unsigned));
	t1 = (unsigned *) calloc((unsigned) parm->len, sizeof(unsigned));

	table = (unsigned **) calloc((unsigned)parm->len / 2, sizeof(unsigned *));
	for (i = 0; i < parm->len / 2; i++)
		table[i] = (unsigned *) calloc((unsigned)2 * parm->len, sizeof(unsigned));

	pshift = bitsfree(key->p, parm->len / 2 - 1);
	bsl(p1, key->p, parm->len / 2 - 1, pshift);
	qshift = bitsfree(key->q, parm->len / 2 - 1);
	bsl(q1, key->q, parm->len / 2 - 1, qshift);

	bsl(b1, in, parm->len - 1, pshift);
	bsl(b2, in, parm->len - 1, qshift);

	fullreduce(b1, parm->len / 2, p1, parm->len / 2 - 1);
	fullreduce(b2, parm->len / 2, q1, parm->len / 2 - 1);

	/* Decrypt using library routine. */

	chexp(r, in, key->n,
	      key->p, p1, pshift,
	      key->q, q1, qshift,
	      key->d1, key->d2, key->a,
	      parm->len - 1, parm->len / 2 - 1,
	      b1, b2, u, v, t1, table);

	for (i = 0; i < parm->len / 2; i++)
		cfree((char*)(table[i]));

	cfree((char*)table);
	cfree((char*)p1);
	cfree((char*)q1);
	cfree((char*)u);
	cfree((char*)v);
	cfree((char*)b1);
	cfree((char*)b2);
	cfree((char*)t1);

	(*out) = r;

	return (OK);
}

int 
rsa_get_blocksize(parm)
	struct RSAParameters *parm;
{
	return BITS * (parm->len);
}

int 
rsa_pack_encrypted(in, out, len, parm)
	unsigned       *in;
	char          **out;
	int            *len;
	struct RSAParameters *parm;
{
	char           *buff;
	unsigned        val;
	char           *cp;
	int             i;
	int             j;

	buff = (char *) malloc((unsigned)parm->len * BITS / 8);
	if (buff == (char *) 0)
		return (NOTOK);

	cp = buff;
	for (i = parm->len - 1; i >= 0; i--) {
		val = in[i];
		for (j = 0; j < 2; j++) {
			*cp = (val >> 8) & 255;
			cp++;
			val = (val << 8) & MASK;
		}
	}

	(*out) = buff;
	(*len) = parm->len * BITS;
/*
	(void)fprintf(stderr, "BITS = %d, len = %d\n", BITS, parm->len);
*/
	return (OK);
}

int 
rsa_unpack_encrypted(in, out, parm)
	char           *in;
	unsigned      **out;
	struct RSAParameters *parm;
{
	int             i;
	int             j;
	int             last;
	char           *cp;
	unsigned       *buff;
	unsigned        val;

	buff = (unsigned *) calloc((unsigned)parm->len, sizeof(unsigned));
	if (buff == NULLENCRYPTED)
		return (NOTOK);

	cp = in;
	last = parm->len - 1;
	if (parm->shift > 7) {
		buff[last] = *cp & 255;
		cp++;
		last--;
	}
	for (i = last; i >= 0; i--) {
		val = 0;
		for (j = 0; j < 2; j++) {
			val = (val << 8) | (*cp & 255);
			cp++;
		}
		buff[i] = val;
	}

	(*out) = buff;

	return (OK);
}

int 
rsa_pack_public_key(in, out, length, parm)
	struct RSAPublicKey *in;
	char          **out;
	int            *length;
	struct RSAParameters *parm;
{
	char           *buff;
	char            ibuff[8];	/* This must be large enough to hold
					 * a unsigned int */
	unsigned       *modulus;
	int             i;
	int             len;
	int             klen;
	unsigned        expo;

	expo = in->e;
	len = 0;
	while (expo != 0) {
		ibuff[len] = expo & 255;
		expo = expo >> 8;
		len++;
	}

	klen = parm->len * 2;

	/* The encoding will be  of length len + klen + 6 (for tagging) */

	buff = (char *) malloc((unsigned)(6 + len + klen));

	if (buff == (char *) 0)
		return (NOTOK);

	buff[0] = (char) 0x30;
	buff[1] = (char) 4 + len + klen;
	buff[2] = (char) 0x02;
	buff[3] = (char) klen;

	modulus = in->n;

	for (i = 0; i < parm->len; i++) {
		buff[4 + 2 * i] = (modulus[parm->len - i - 1] >> 8) & 255;
		buff[5 + 2 * i] = (modulus[parm->len - i - 1]) & 255;
	}

	expo = in->e;

	buff[4 + klen] = 0x02;
	buff[5 + klen] = len;
	for (i = 0; i < len; i++)
		buff[6 + klen + i] = ibuff[len - i - 1];

	*out = buff;
	*length = len + klen + 6;

	return (OK);
}

int 
rsa_unpack_public_key(in, out, parm)
	char           *in;
	struct RSAPublicKey **out;
	struct RSAParameters **parm;
{
	struct RSAPublicKey *result;
	char           *buff;
	int             len;
	int             i;
	unsigned       *modulus;
	unsigned        word;
	struct RSAParameters *n_parm;

	/* Sanity checks */

	if (in[0] != 0x30)
		return (NOTOK);

	if (in[2] != 0x02)
		return (NOTOK);

	if (in[3] != 0x40)
		return (NOTOK);	/* Unexpected  key size */

	if (in[4 + 64] != 0x02)
		return (NOTOK);

	if (in[5 + 64] > 4)
		return (NOTOK);	/* Unexpected exponent size */

	n_parm = (struct RSAParameters *)
		calloc((unsigned)1, sizeof(*n_parm));

	if (n_parm == (struct RSAParameters *) 0)
		return (NOTOK);

	/*
	 * HACK: assume key length is 512 bits
	 */

	n_parm->len = KEYSIZE;

	result = (struct RSAPublicKey *) calloc((unsigned)1, sizeof(*result));
	if (result == (struct RSAPublicKey *) 0)
		return (NOTOK);

	modulus = (unsigned *) calloc((unsigned)n_parm->len, sizeof(unsigned));
	result->n = modulus;

	buff = in;

	for (i = 0; i < n_parm->len; i++) {
		word = ((255 & buff[4 + 2 * i]) << 8) | (255 & buff[5 + 2 * i]);
		modulus[(KEYSIZE - i) - 1] = word;
	}

	len = buff[5 + n_parm->len * 2];
	word = 0;
	for (i = 0; i < len; i++) {
		word = word << 8;
		word |= (255 & buff[6 + n_parm->len * 2 + i]);
	}
	result->e = word;

	*parm = n_parm;
	*out = result;

	return (OK);
}

int 
rsa_pack_secret_key(in, out, length, parm)
	struct RSASecretKey *in;
	char          **out;
	int            *length;
	struct RSAParameters *parm;
{
	char           *buff;
	int             expo;
	int             elen;
	int             klen;
	int             len;
	int             i;
	unsigned       *num;
	char            ibuff[8];	/* Must be able to hold a long */

	/*
	 * To reconstruct the secret key, only p, q and e are needed. The
	 * secret key is transmitted as the ASN.1 structure :
	 * 
	 * SEQUENCE { p INTEGER, q INTEGER, e INTEGER}
	 */

	expo = 65537;		/* dont have support for other values yet */
	elen = 0;
	while (expo != 0) {
		ibuff[elen] = expo & 255;
		expo = expo >> 8;
		elen++;
	}

	/*
	 * The RSA routines pack 16 bits to an unsigned. The length of the
	 * factors is half that of the modulus. Hence factors take parm->len
	 * bytes. WARNING: This will not apply on all m/c architectures.
	 */

	klen = parm->len;	/* Length of a factor of n (in bytes) */

	len = klen * 2 + elen + 8;	/* Two factors, the exponent, and 4
					 * tags */

	buff = (char *) malloc((unsigned)len);

	if (buff == (char *) 0)
		return (NOTOK);

	buff[0] = 0x30;
	buff[1] = len - 2;
	buff[2] = 0x02;
	buff[3] = klen;

	num = in->p;
	for (i = 0; i < klen / 2; i++) {
		buff[4 + 2 * i] = (num[klen / 2 - i - 1] >> 8) & 255;
		buff[5 + 2 * i] = (num[klen / 2 - i - 1]) & 255;
	}

	buff[4 + klen] = 0x02;
	buff[5 + klen] = klen;

	num = in->q;
	for (i = 0; i < klen / 2; i++) {
		buff[6 + klen + 2 * i] = (num[klen / 2 - i - 1] >> 8) & 255;
		buff[7 + klen + 2 * i] = (num[klen / 2 - i - 1]) & 255;
	}

	buff[6 + klen * 2] = 0x02;
	buff[7 + klen * 2] = elen;
	for (i = 0; i < elen; i++)
		buff[8 + klen * 2 + i] = ibuff[elen - i - 1];

	*out = buff;
	*length = len;

	return (OK);
}

int 
rsa_unpack_secret_key(in, out, parm)
	char           *in;
	struct RSASecretKey **out;
	struct RSAParameters **parm;
{
	struct RSASecretKey *secret;
	struct RSAParameters *n_parms;
	char           *buff;
	unsigned       *e;
	unsigned       *num;
	unsigned        word;
	int             klen;
	int             i;

	/*
	 * This RSA implementation pre-calcualtes certain values. When a
	 * secret key is unpacked, these must be calculated. (The secret key
	 * could have come from an implementation which never needed to
	 * compute these values).
	 */

	n_parms = (struct RSAParameters *) calloc((unsigned)1, sizeof(*n_parms));

	if (n_parms == (struct RSAParameters *) 0)
		return (NOTOK);


	/* HACK: assume keysize is 512 bits */
	klen = KEYSIZE;		/* length of a factor (in bytes) */
	n_parms->len = KEYSIZE;	/* length of the modulus (in zoned words) */

	secret = (struct RSASecretKey *) calloc((unsigned)1, sizeof(*secret));
	if (secret == (struct RSASecretKey *) 0)
		return (NOTOK);

	secret->p = (unsigned *) calloc((unsigned)n_parms->len / 2, sizeof(unsigned));
	secret->q = (unsigned *) calloc((unsigned)n_parms->len / 2, sizeof(unsigned));
	secret->n = (unsigned *) calloc((unsigned)n_parms->len, sizeof(unsigned));
	secret->a = (unsigned *) calloc((unsigned)n_parms->len, sizeof(unsigned));
	secret->d = (unsigned *) calloc((unsigned)n_parms->len, sizeof(unsigned));
	secret->d1 = (unsigned *) calloc((unsigned)n_parms->len / 2, sizeof(unsigned));
	secret->d2 = (unsigned *) calloc((unsigned)n_parms->len / 2, sizeof(unsigned));

	e = (unsigned *) calloc((unsigned)n_parms->len, sizeof(unsigned));
	e[0] = 1;
	e[1] = 1;		/* 65537 again! */

	buff = in;
/*
	printf("key length = %d\n", (unsigned) buff[3]);
*/
	num = secret->p;

	for (i = 0; i < klen / 2; i++) {
		word = ((255 & buff[4 + 2 * i]) << 8) | (255 & buff[5 + 2 * i]);
		num[klen / 2 - i - 1] = word;
	}

	num = secret->q;

	for (i = 0; i < klen / 2; i++) {
		word = ((255 & buff[6 + klen + 2 * i]) << 8) | (255 & buff[7 + klen + 2 * i]);
		num[klen / 2 - i - 1] = word;
	}

	rsaparms(secret->p, secret->q, secret->n,
		 e, secret->d, secret->a,
		 secret->d1, secret->d2,
		 n_parms->len / 2 - 1, n_parms->len - 1);

	*parm = n_parms;
	*out = secret;

	return (OK);
}

  
static 
gulp(f)
	FILE           *f;
{
	int             i;

	for (i = 0; i < 5; i++)
		(void) getc(f);
}

  
static 
hexin(f, a, l)
	FILE           *f;
	unsigned        a[];
	int             l;
{
	int             i;
	int             j;
	int             n;
	char            c;

	for (i = l; i >= 0; i--) {
		a[i] = 0;
		for (j = 0; j < NPW; j++) {
			c = getc(f);
			n = (int) c - (int) '0';
			if (c > '9')
				n = (int) c - (int) 'a' + 10;
			a[i] = (a[i] << 4) + n;
		}
	}
	(void)fscanf(f, " \n");
}

int 
rsa_write_public_key(in, out, parms)
	struct RSAPublicKey *in;
	char           *out;
	struct RSAParameters *parms;
{
	FILE           *f1;

	f1 = fopen(out, "w");
	if (f1 == NULL)
		return (NOTOK);

	(void)fprintf(f1, "L  = %d\n", parms->len * 16 - parms->shift);
	(void)fprintf(f1, "N  = ");
	fhexdump(f1, in->n, parms->len - 1);
	(void)fprintf(f1, "E  = %x\n", in->e);

	(void)fclose(f1);

	return (OK);
}


int 
rsa_read_public_key(in, out, parms)
	char           *in;
	struct RSAPublicKey **out;
	struct RSAParameters **parms;
{
	FILE           *f1;
	struct RSAPublicKey *result;
	struct RSAParameters *newparms;

	result = (struct RSAPublicKey *) calloc((unsigned)1, sizeof(*result));
	if (result == (struct RSAPublicKey *) 0)
		return (NOTOK);

	newparms = (struct RSAParameters *) calloc((unsigned)1, sizeof(*newparms));
	if (newparms == (struct RSAParameters *) 0)
		return (NOTOK);

	f1 = fopen(in, "r");
	if (f1 == NULL)
		return (NOTOK);

	(void)fscanf(f1, "L  = %d\n", &(newparms->len));
	newparms->shift = 16 - newparms->len % 16;
	if (newparms->shift == 16)
		newparms->shift = 0;
	newparms->len = (newparms->len + 15) / 16;

	result->n = (unsigned *) calloc((unsigned)newparms->len, sizeof(unsigned));
	if (result->n == NULLENCRYPTED)
		return (NOTOK);

	gulp(f1);
	hexin(f1, result->n, newparms->len - 1);
	gulp(f1);
	(void)fscanf(f1, "%x", &(result->e));
	(void)fscanf(f1, " \n");

	(void)fclose(f1);

	(*out) = result;
	(*parms) = newparms;

	return (OK);
}

int 
rsa_write_secret_key(in, out, parm)
	struct RSASecretKey *in;
	char           *out;
	struct RSAParameters *parm;
{
	FILE           *f1;

	f1 = fopen(out, "w");
	if (f1 == NULL)
		return (NOTOK);

	(void)fprintf(f1, "L  = %d\n", parm->len * 16 - parm->shift);
	(void)fprintf(f1, "P  = ");
	fhexdump(f1, in->p, parm->len / 2 - 1);
	(void)fprintf(f1, "Q  = ");
	fhexdump(f1, in->q, parm->len / 2 - 1);
	(void)fprintf(f1, "A  = ");
	fhexdump(f1, in->a, parm->len / 2 - 1);
	(void)fprintf(f1, "N  = ");
	fhexdump(f1, in->n, parm->len - 1);
	(void)fprintf(f1, "D  = ");
	fhexdump(f1, in->d, parm->len - 1);
	(void)fprintf(f1, "d1 = ");
	fhexdump(f1, in->d1, parm->len / 2 - 1);
	(void)fprintf(f1, "d2 = ");
	fhexdump(f1, in->d2, parm->len / 2 - 1);

	(void)fclose(f1);

	return (OK);
}

int 
rsa_read_secret_key(in, out, parm)
	char           *in;
	struct RSASecretKey **out;
	struct RSAParameters **parm;
{
	FILE           *f1;
	struct RSASecretKey *key;
	struct RSAParameters *newparm;
	int             keysize;

	f1 = fopen(in, "r");
	if (f1 == NULL)
		return (NOTOK);

	newparm = (struct RSAParameters *) calloc((unsigned)1, sizeof(*newparm));

	if (newparm == (struct RSAParameters *) 0)
		return (NOTOK);

	(void)fscanf(f1, "L  = %d\n", &keysize);

	newparm->shift = 16 - (keysize  % 16);
	if (newparm->shift == 16)
		newparm->shift = 0;
	keysize = (keysize+15)/16;
	newparm->len = keysize;

	key = (struct RSASecretKey *) calloc((unsigned)1, sizeof(*key));

	if (key == (struct RSASecretKey *) 0)
		return (NOTOK);

	key->p = (unsigned *) calloc((unsigned)keysize / 2, sizeof(unsigned));
	if (key->p == NULLENCRYPTED)
		return (NOTOK);

	key->q = (unsigned *) calloc((unsigned)keysize / 2, sizeof(unsigned));
	if (key->q == NULLENCRYPTED)
		return (NOTOK);

	key->a = (unsigned *) calloc((unsigned)keysize / 2, sizeof(unsigned));
	if (key->a == NULLENCRYPTED)
		return (NOTOK);

	key->n = (unsigned *) calloc((unsigned)keysize, sizeof(unsigned));
	if (key->n == NULLENCRYPTED)
		return (NOTOK);

	key->d = (unsigned *) calloc((unsigned)keysize, sizeof(unsigned));
	if (key->d == NULLENCRYPTED)
		return (NOTOK);

	key->d1 = (unsigned *) calloc((unsigned)keysize / 2, sizeof(unsigned));
	if (key->d1 == NULLENCRYPTED)
		return (NOTOK);

	key->d2 = (unsigned *) calloc((unsigned)keysize / 2, sizeof(unsigned));
	if (key->d2 == NULLENCRYPTED)
		return (NOTOK);

	gulp(f1);
	hexin(f1, key->p, keysize / 2 - 1);
	gulp(f1);
	hexin(f1, key->q, keysize / 2 - 1);
	gulp(f1);
	hexin(f1, key->a, keysize / 2 - 1);
	gulp(f1);
	hexin(f1, key->n, keysize - 1);
	gulp(f1);
	hexin(f1, key->d, keysize - 1);
	gulp(f1);
	hexin(f1, key->d1, keysize / 2 - 1);
	gulp(f1);
	hexin(f1, key->d2, keysize / 2 - 1);

	(void)fclose(f1);

	(*out) = key;
	(*parm) = newparm;

	return (OK);
}

/* ARGSUSED */
int 
rsa_free_encrypted(enc, parm)
	unsigned       *enc;
	struct RSAParameters *parm;
{
/*	free((char*)parm);	*/
	free((char*)enc);

	return (OK);
}

int 
rsa_free_public_key(key)
	struct RSAPublicKey *key;
{
	free((char*)key->n);
	free((char*)key);

	return (OK);
}

int 
rsa_free_secret_key(key)
	struct RSASecretKey *key;
{
	free((char*)key->n);
	free((char*)key->p);
	free((char*)key->q);
	free((char*)key->d1);
	free((char*)key->d2);
	free((char*)key->d);
	free((char*)key->a);
	free((char*)key);

	return (OK);
}

int 
rsa_secret_to_public(in, out, parms)
	struct RSASecretKey *in;
	struct RSAPublicKey **out;
	struct RSAParameters *parms;
{
	struct RSAPublicKey *result;

	result = (struct RSAPublicKey *) calloc((unsigned)1, sizeof(*result));
	if (result == (struct RSAPublicKey *) 0)
		return (NOTOK);

	result->e = 65537;
	result->n = (unsigned *) calloc((unsigned)parms->len, sizeof(unsigned));
	bcopy((char*)(in->n), (char*)(result->n), parms->len * sizeof(unsigned));

	*out = result;

	return (OK);
}

int 
rsa_compare_encrypted(a, b, parms)
	unsigned       *a;
	unsigned       *b;
	struct RSAParameters *parms;
{
	int             result, i;

	result = OK;

	for (i = 0; i < parms->len; i++)
		if (a[i] != b[i])
			result = NOTOK;

	return (result);
}
